/**
 Use raw I/O calls to extract information about Windows
  "PE" executable (EXE) files.  This program uses simple
  binary I/O, and hence should run on any machine.
 
 Orion Sky Lawlor, olawlor@acm.org, 2005/3/31 (Public Domain)
*/
#include <stdio.h>
#include <iostream>
#include <string>
using namespace std;

void dumpEXE(const char *exeFile);

int main(int argc,char *argv[])
{
	if (argc>1) {dumpEXE(argv[1]); return 0;}
	cout << "Usage: dumpEXE <exe file name>" << endl;
	return 0;
}

/**
 Data types for binary I/O.
*/
typedef unsigned char byte;

/**
 A truly weird class: in memory, this class is 
 always a little-endian 32-bit integer.  Hence on 
 an x86 machine, it's the same as just "int".
 But it has the exact same layout on any other machine,
 like a 64-bit big-endian machine, where "int" 
 is 8 bytes, and the bytes are in the wrong order.
 This class also has no alignment restrictions, unlike "int",
 which the compiler may add padding around.
*/
class lil32 {
private:
	/** Storage for the 32-bit integer we describe. */
	byte v[4];
public:
	/// "Cast-to-int" operator.
	///  This operator allows us to read from "lil32" objects
	///  just like they were regular "int"s.
	///  It reassembles the integer value from our bytes.
	inline operator int () const {
		return v[0]+(v[1]<<8)+(v[2]<<16)+(v[3]<<24);
	}
};

/** Like lil32, but for a 16-bit little-endian integer. */
class lil16 {
private:
	/** Storage for the 16-bit integer we describe. */
	byte v[2];
public:
	/// "Cast-to-int" operator.
	inline operator int () const {
		return v[0]+(v[1]<<8);
	}
};


/**
 A header stored in the guts of every windows EXE file.
 This information comes from the Microsoft documentation,
 for example, at http://msdn.microsoft.com/library/default.asp?url=/library/en-us/dndebug/html/msdn_peeringpe.asp
 
 It's substantially identical to the Microsoft
 "IMAGE_FILE_HEADER" and "IMAGE_OPTIONAL_HEADER" structures.
*/
struct myPEheader {
	byte pe[2]; /* 'P' 'E' */
	byte dunno1[22]; /* random unknown stuff */
	byte magic[2]; /* 0x0b, 0x01 */
	byte MajorLinkerVersion;
	byte MinorLinkerVersion;
	/* Sizes (byte counts) */
	lil32 SizeOfCode;
	lil32 SizeOfInitializedData;
	lil32 SizeOfUninitializedData;
	/* Pointers (to place in memory to load data) */
	lil32 AddressOfEntryPoint;
	lil32 BaseOfCode;
	lil32 BaseOfData;
	lil32 ImageBase;
	/* Really useless stuff */
	lil32 SectionAlignment;
	lil32 FileAlignment;
	lil16 MajorOperatingSystemVersion;
	lil16 MinorOperatingSystemVersion;
	lil16 MajorImageVersion;
	lil16 MinorImageVersion;
	lil16 MajorSubsystemVersion;
	lil16 MinorSubsystemVersion;
	lil32 Reserved1;
	lil32 SizeOfImage;
	lil32 SizeOfHeaders;
	lil32 CheckSum; /* normally 0 */
	lil32 Subsystem; /* 2- GUI, 3- console, 5- OS/2, 7-posix */
	/* and lots of other stuff */
};


/// Print this integer size field.
void printSize(const char *what,int bytes) {
	cout<<"Size of "<<what<<": "<<
		dec<<bytes<<" bytes "<<
		"(0x"<<hex<<bytes<<dec<<")\n";
}

/// Print this address.
void printAddr(const char *what,void *ptr) {
	cout<<"Address of "<<what<<":  "<<ptr<<"\n";
}

/// Print summary information about this file to the screen.
void dumpEXE(const char *exeFile)
{
	cout<<"Image name: '"<<exeFile<<"'\n";

/* Open the .exe file for binary read: */
	FILE *f=fopen(exeFile,"rb");
	if (f==NULL) {printf("Can't open executable!\n");exit(1);}
	
/* Jump to, and read, the offset to the PE header.
  We do this by seeking into the file at the appropriate
  spot, and reading one 32-bit little-endian integer.
*/
	lil32 peOffset; /* offset from the start of file to PE header */

	fseek(f,0x3c,SEEK_SET); 
	fread(&peOffset,sizeof(peOffset),1,f);
	if (peOffset<0x10 || peOffset>0x1000) {printf("Bogus PE offset!\n");exit(1);}
	cout<<"File offset to PE record: "<<peOffset<<" bytes\n";
	
/* Read in the PE header */
	myPEheader *h=new myPEheader;
	fseek(f,peOffset,SEEK_SET);
	fread(h,sizeof(myPEheader),1,f);
	if (h->pe[0]!='P' || h->pe[1]!='E') {printf("Bogus PE header!\n");exit(1);}
	
/* Print out stuff from the header */
	printSize("code",h->SizeOfCode);
	printSize("initialized data",h->SizeOfInitializedData);
	printSize("uninitialized data",h->SizeOfUninitializedData);
	printAddr("code",(void *)(int)h->BaseOfCode);
	printAddr("code entry point",(void *)(int)h->AddressOfEntryPoint);
	printAddr("data",(void *)(int)h->BaseOfData);
	printAddr("image",(void *)(int)h->ImageBase);
	cout<<"Linker version: "<<(int)h->MajorLinkerVersion<<"."<<(int)h->MinorLinkerVersion<<"\n";
	
	const static char *subsystem_names[]={
		"0-- unknown",
		"1-- 'NATIVE' (null) subsystem",
		"2-- windows GUI",
		"3-- windows command line",
		"4-- unknown!",
		"5-- OS/2 command line",
		"6-- unknown",
		"7-- POSIX command line",
		"8-- unknown"
	};
	if (h->Subsystem<=8)
		cout<<"Subsystem: "<<subsystem_names[h->Subsystem]<<"\n";
	else
		cout<<"Subsystem: "<<h->Subsystem<<": unknown.\n";

/* Clean up afterwards */
	delete h;
	fclose(f);
}
